import { Account, ConnectConfig, Near, keyStores } from 'near-api-js';
import { Command, Flags } from '@oclif/core';
import { FinalExecutionOutcome } from 'near-api-js/lib/providers';
import { OutputFlags } from '@oclif/core/lib/interfaces';
import { homedir } from 'os';

import BN from 'bn.js';
import { Tonic } from '@tonic-foundation/tonic';

import {
  getNearConfig,
  NearEnv
} from '@tonic-foundation/config';
import { getExplorerUrl } from './utils';

const getKeystore = async () => {
  const HOME_DIR = homedir();
  const CREDENTIALS_DIR = '.near-credentials';
  const credentialsPath = require('path').join(HOME_DIR, CREDENTIALS_DIR);

  return new keyStores.UnencryptedFileSystemKeyStore(credentialsPath);
};

export default abstract class BaseCommand extends Command {
  static flags = {
    networkId: Flags.string({
      default: 'testnet',
      env: 'NEAR_ENV',
      options: ['mainnet', 'testnet', 'localnet']
    }),
    accountId: Flags.string({
      env: 'NEAR_ACCOUNT_ID',
      required: true,
    }),
    contractId: Flags.string({
      env: 'TONIC_CONTRACT_ID',
      required: true,
    }),
  };

  account!: Account;

  globalFlags!: OutputFlags<typeof BaseCommand.flags>;

  tonic!: Tonic;

  connectConfig!: ConnectConfig;

  async init() {
    const { flags } = await this.parse(this.constructor as typeof BaseCommand);
    this.globalFlags = flags;

    const connectConfig = getNearConfig(flags.networkId as NearEnv);
    this.connectConfig = connectConfig;

    const keyStore = await getKeystore();
    const near = new Near({ ...connectConfig, keyStore });
    const account = await near.account(flags.accountId);
    this.account = account;

    this.tonic = new Tonic(account, flags.contractId);
  }

  logTransaction(outcome: FinalExecutionOutcome) {
    const txId = outcome.transaction_outcome.id;
    this.log(`Transaction ID: ${txId}`);
    const explorerLink = `${getExplorerUrl(this.connectConfig.networkId as NearEnv)}/${txId}`;
    this.log(`View in the explorer: ${explorerLink}`);
  }
}

export const BNflag = Flags.build<BN>({
  parse: async (s) => new BN(s),
});

export const FloatFlag = Flags.build<number>({
  parse: async (s) => {
    const v = parseFloat(s);
    if (isNaN(v)) {
      throw new Error(`${s} is not a number`);
    }
    return v;
  },
});

export const TokenIdFlag = Flags.build<string>({
  parse: parseTokenId,
});

export async function parseTokenId(s: string): Promise<string> {
  if (s.toUpperCase() === 'NEAR') {
    return 'NEAR';
  }
  return s;
}
